// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <math.h>
#include <gsl/gsl_multimin.h>
#define SWAP(a,b) tempr=(a);(a)=(b);(b)=tempr

struct istruct main_g;
struct istruct main_exp;
struct istruct main_error;
struct istruct main_fit;
static double t0=4e-6;
static double v0=1e5;
static double sigma=1e-20;

int fft_len=200;
//Replaces data[1..2*nn] by its discrete Fourier transform, if isign is input as 1; or replaces
//data[1..2*nn] by nn times its inverse discrete Fourier transform, if isign is input as −1.
//data is a complex array of length nn or, equivalently, a real array of length 2*nn. nn MUST
//be an integer power of 2 (this is not checked for!).
static double error_tot=0.0;

double sech(double x)
{
return 2.0/(exp(x)+exp(-x));
}

double get_fft_error (const gsl_vector *v, void *params)
{
int n;
for (n=0;n<main_g.len;n++)
{
	main_g.data[n]=fabs(gsl_vector_get(v, n));
}
t0=fabs(gsl_vector_get(v, n));
int i;
int ii;


double En=0.0;
double Ep=0.0;
double E=0.0;
double dE=0.0;
double Eomega=0.0;
double sum=0.0;

for (ii=1;ii<main_exp.len;ii++)
{

Eomega=main_exp.x[ii];
sum=0.0;
	for (i=1;i<main_g.len;i++)
	{

		if (i==1)
		{
			En=main_g.x[i];
		}else
		{
			En=main_g.x[i-1];
		}

		if (i==main_g.len-1)
		{
			Ep=main_g.x[i];
		}else
		{
			Ep=main_g.x[i+1];
		}

		E=main_g.x[i];
		dE=fabs((Ep-En)/2.0);

		sum+=(sigma*v0/2.0)*main_g.data[i]*sech(q*(Eomega-E)/(300.0*kb))*dE;
	}

main_fit.x[ii]=main_exp.x[ii];
main_fit.data[ii]=sum;

main_error.x[ii]=main_exp.x[ii];
main_error.data[ii]=fabs(main_exp.data[ii]-main_fit.data[ii]);
}
error_tot=0.0;
for (i=0;i<main_exp.len;i++)
{
if (main_error.x[i]<-0.28)
{
error_tot+=main_error.data[i];
}
}

inter_save_a(&main_exp,"./","rod_exp.dat");
inter_save_a(&main_fit,"./","rod_fit.dat");
inter_save_a(&main_error,"./","rod_error.dat");
//printf("Dump\n");
//getchar();
return error_tot;
}

double fft_mag(struct istruct* in,int pos)
{
return sqrt(in->data[pos*2]*in->data[pos*2]+in->data[(pos*2)+1]*in->data[(pos*2)+1]);
}



void fft_extract()
{

/*

int i;
struct istruct data;
struct istruct dos;

inter_load_a(&data,"./exp/tof0/bump.dat",0,1);
inter_copy(&dos,&data);
double v0=0.0;
double sigma=1e-20;
double v=1e5;
double N0=1e27;
double t0=1e-7;
double Q0=1e30;
double E=0.0;
double t=0.0;
for (i=0;i<data.len;i++)
{
	t=dos.x[i];
	v0=1/t0;//sigma*v*N0*kb*300.0;
	E=kb*300.0*log(v0*t);
	dos.x[i]=E/1.6e-19;
	dos.data[i]=data.data[i]*t*2.0/(t0*Q0*v0);
	inter_save_a(&dos,"./","dosout.dat");
}

inter_free(&data);
*/


int i;
double phase=0.0;
double mag=0.0;
struct istruct fft;
struct istruct fft_in;
struct istruct data;

struct istruct fft_mag;
struct istruct fft_phase;
struct istruct fft_mag_small;
struct istruct fft_phase_small;
struct istruct dos;

//fft_new(&fft);
//fft_new(&fft_in);
inter_load_a(&data,"./exp/tof0/bump.dat",0,1);
double t=0.0;
double E=0.0;

double N0=1e27;
double t0=1e-4;
double Q0=1e27;

inter_copy(&dos,&data,TRUE);
for (i=0;i<data.len;i++)
{
	t=dos.x[i];
	v0=1/t0;//sigma*v*N0*kb*300.0;
	E=kb*300.0*log(v0*t);
	dos.x[i]=E/1.6e-19;
	dos.data[i]=data.data[i]*t*2.0/(t0*Q0*v0);

}
	inter_save_a(&dos,"./","dosout.dat");
//inter_new(&data,fft_len);
/*double pos=0.0;
for (i=0;i<data.len;i++)
{
	data.x[i]=pos;
	data.data[i]=sin(2.0*pi*((double)i/(double)data.len));
	pos+=1.0;
}*/

//inter_save_a(&data,"./","sin.dat");


inter_new(&fft,fft_len*2);
inter_new(&fft_in,data.len*2);


i_to_fft(&fft_in,&data);

//do_fft(&fft_in, 1);
//inter_copy(&fft,&fft_in,FALSE);
do_dft(&fft,&fft_in);




inter_new(&fft_mag,fft_len);
fft_to_mag(&fft,&fft_mag);

inter_new(&fft_phase,fft_len);
fft_to_phase(&fft,&fft_phase);

for (i=1;i<fft_mag.len;i++)
{
fft_mag.x[i]=kb*300.0*log(fft_mag.x[i]*t0)/1.6e-19;
fft_phase.x[i]=fft_mag.x[i];
}
//inter_chop(&fft_mag,0.05, 0.18);
//inter_chop(&fft_phase,0.05, 0.18);

inter_new(&main_g,fft_mag.len);
inter_new(&main_fit,fft_mag.len);
inter_new(&main_error,fft_mag.len);
inter_new(&main_exp,fft_mag.len);

inter_save_a(&fft_mag,"./","fft_mag.dat");
inter_save_a(&fft_phase,"./","fft_phase.dat");


for (i=1;i<fft_mag.len;i++)
{
	phase=fft_phase.data[i];
	mag=fft_mag.data[i];

	main_exp.data[i]=((data.data[0]*sin(phase))/(mag))-fft_mag.x[i];
	main_g.data[i]=main_exp.data[i]*1e16;
	main_exp.x[i]=fft_mag.x[i];
	main_g.x[i]=fft_mag.x[i];
}


inter_save_a(&main_exp,"./","exp.dat");

int fitvars=main_g.len+1;
size_t iter = 0;
int status;
double size;
double shift=1.0;
double par[5] = {1.0, 2.0, 10.0, 20.0, 30.0};
const gsl_multimin_fminimizer_type *T = gsl_multimin_fminimizer_nmsimplex;

gsl_multimin_fminimizer *s = NULL;
gsl_vector *ss, *x;
gsl_multimin_function minex_func;



////////////////////////////////////////////////////////////
x = gsl_vector_alloc (fitvars);
ss = gsl_vector_alloc (fitvars);

int item=0;
int n;

for (n=0;n<main_g.len;n++)
{
	gsl_vector_set (x, item, main_g.data[n]);
	gsl_vector_set (ss, item, fabs(main_g.data[n]*shift));
	item++;
}
	gsl_vector_set (x, item,t0);
	gsl_vector_set (ss, item, t0);
	item++;

printf("rod =%d\n",item);
minex_func.n = item;
minex_func.f = get_fft_error;
minex_func.params = par;

s = gsl_multimin_fminimizer_alloc (T, item);
gsl_multimin_fminimizer_set (s, &minex_func, x, ss);
int loop=0;

do
{
	iter++;
	status = gsl_multimin_fminimizer_iterate(s);

	if (status) break;

	size = gsl_multimin_fminimizer_size (s);
	status = gsl_multimin_test_size (size, 1e-10);

	//if (status == GSL_SUCCESS)
	//{
	//	printf ("converged to minimum at\n");
	//}

	for (n=0;n<main_g.len;n++)
	{
		main_g.data[n]=fabs(gsl_vector_get(s->x, n));
		//printf("%d %le %le %le\n",n,main_g.data[n],main_fit.data[n],main_exp.data[n]);
	}

	//inter_smooth(&main_g,5);
	//inter_smooth(&main_g,5);

	inter_save_a(&main_g,"./","g.dat");

	printf("%le\n",error_tot);
	loop++;

	if (loop>1000)
	{
		//getchar();
		loop=0;
	}


}while (status == GSL_CONTINUE);


gsl_vector_free(x);
gsl_vector_free(ss);
gsl_multimin_fminimizer_free (s);


/////////////////////////////////////////////////////////////





inter_save_a(&main_fit,"./","fit.dat");
inter_save_a(&main_error,"./","error.dat");

inter_free(&fft_mag);
inter_free(&fft_phase);
inter_free(&fft_mag_small);
inter_free(&fft_phase_small);

inter_free(&data);
inter_free(&fft);
inter_free(&main_g);
inter_free(&main_fit);
inter_free(&main_exp);
inter_free(&main_error);
inter_free(&fft_in);
}

void fft_new(struct istruct* in)
{
inter_new(in,fft_len*2);
}

void fft_to_mag(struct istruct* in,struct istruct* out)
{

int i;
for (i=0;i<in->len/2;i++)
{
	//printf("%le\n",in->data[i*2]);
	//getchar();
	out->data[i]=fft_mag(in,i);

	out->x[i]=in->x[i*2];
}


}

void fft_to_phase(struct istruct* in,struct istruct* out)
{

int i;
for (i=0;i<in->len/2;i++)
{
	out->data[i]=atan2(in->data[(i*2)+1],in->data[(i*2)]);
	out->x[i]=in->x[i*2];
}


}

void fft_to_real(struct istruct* in,struct istruct* out)
{

int i;
for (i=0;i<in->len/2;i++)
{
	out->data[i]=in->data[i*2];
	out->x[i]=in->x[i*2];
}


}

void i_to_fft(struct istruct* out,struct istruct* in)
{
int i=0;
for (i=0;i<in->len;i++)
{
	out->data[i*2]=in->data[i];
	out->x[i*2]=in->x[i];
	out->x[(i*2)+1]=in->x[i];
}

}

void do_dft(struct istruct* out,struct istruct* in)
{
int i;
int ii;
double tn=0.0;
double tp=0.0;
double dt=0.0;
double t=0.0;
double omega=0.0;
double a=0.0;
double b=0.0;

double delta=(in->x[((in->len/2-1)*2)]-in->x[(0*2)])/(in->len/2);//in->x[(1*2)]-in->x[(0*2)];//
for (i=0;i<out->len/2;i++)
{
out->data[i*2]=0.0;
out->data[(i*2)+1]=0.0;
omega=2.0*pi*((double)i/(double)(out->len/2))/delta;
//printf("%le",omega);
//getchar();
//if (i==0) omega=0.0;
	for (ii=0;ii<in->len/2;ii++)
	{
		if (ii==0)
		{
			tn=0.0;
		}else
		{
			tn=in->x[(ii*2)-1];
		}

		if (ii==in->len-1)
		{
			tp=in->x[(ii*2)]+delta/2.0;
		}else
		{
			tp=in->x[(ii*2)+1];
		}

		dt=(tp-tn)/2.0;
		t=in->x[ii*2];

		a=in->data[ii*2];
		b=in->data[(ii*2)+1];
		//out->data[i*2]+=(a*cos(omega*t)+b*sin(omega*t))*dt;
		//out->data[(i*2)+1]+=(-a*sin(omega*t)+b*cos(omega*t))*dt;
		out->data[i*2]+=(a*cos(omega*t)-b*sin(omega*t))*dt;
		out->data[(i*2)+1]+=(a*sin(omega*t)+b*cos(omega*t))*dt;
		//printf("%le %d %d\n",in->x[i*2],i,ii);
		//getchar();


	}

	out->x[i*2]=omega;
	out->x[(i*2)+1]=omega;

		//printf("%le %le\n",omega,sqrt(out->data[i*2]*out->data[i*2]+out->data[(i*2)+1]*out->data[(i*2)+1]));
		//getchar();
}

}

void do_fft(struct istruct* in, int isign)
{
unsigned long nn=in->len/2;
double *data=in->data-1;
unsigned long n,mmax,m,j,istep,i;
double wtemp,wr,wpr,wpi,wi,theta;
//Double precision for the trigonomet-ric recurrences.
double tempr,tempi;
n=nn << 1;
j=1;
for (i=1;i<n;i+=2)
{
	//This is the bit-reversal section of the
		if (j > i)
		{
			//routine.
			SWAP(data[j],data[i]);
			//Exchange the two complex numbers.
			SWAP(data[j+1],data[i+1]);
		}

		m=n >> 1;
		while (m >= 2 && j > m)
		{
			j -= m;
			m >>= 1;
		}
	j += m;
}
//Here begins the Danielson-Lanczos section of the routine.
mmax=2;
while (n > mmax)
{
	//Outer loop executed log2 nn times.
	istep=mmax << 1;
	theta=isign*(6.28318530717959/mmax);
	//Initialize the trigonometric recurrence.
	wtemp=sin(0.5*theta);
	wpr = -2.0*wtemp*wtemp;
	wpi=sin(theta);
	wr=1.0;
	wi=0.0;
	for (m=1;m<mmax;m+=2)
	{
			for (i=m;i<=n;i+=istep)
			{
				j=i+mmax;
				tempr=wr*data[j]-wi*data[j+1];
				tempi=wr*data[j+1]+wi*data[j];
				data[j]=data[i]-tempr;
				data[j+1]=data[i+1]-tempi;
				data[i] += tempr;
				data[i+1] += tempi;
			}
		wr=(wtemp=wr)*wpr-wi*wpi+wr;
		wi=wi*wpr+wtemp*wpi+wi;
	}
	mmax=istep;
	//Here are the two nested inner loops. This is the Danielson-Lanczos formula: Trigonometric recurrence.
}

double delta=(in->x[(1*2)]-in->x[(0*2)]);
printf("%le %le %le\n",delta,in->x[(1*2)],in->x[(0*2)]);
if (isign==-1)
{
	for (i=1;i<in->len;i++)
	{
	in->data[i]/=(double)fft_len;
	}

}else
{

	for (i=0;i<in->len/2;i++)
	{
		in->x[(i*2)]=2.0*pi*((double)i)/((double)(in->len/2)*delta);
		in->x[(i*2)+1]=2.0*pi*((double)i)/((double)(in->len/2)*delta);
		//printf("%le\n",delta*(double)i);
	}
}

}

