// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <gsl/gsl_multimin.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "sim.h"
#include "server.h"
#include "fit.h"
#include <dirent.h>

void fit_set_light(double in)
{
edit_file("light.inp","",2,in);
}

void fit_set_volt_step(double in)
{
edit_file("jv.inp","",6,in);
}

void fit_set_volt_start(double in)
{
edit_file("jv.inp","",2,in);
}

void fit_set_volt_stop(double in)
{
edit_file("jv.inp","",4,in);
}


void fit_set_sim_mode(int in)
{
edit_file_int("sim.inp","",2,in);
}

void fit_set_external_volt(double in)
{
edit_file("sim.inp","",32,in);
}

void fit_set_tpc_externalv(double in)
{
edit_file("tpc.inp","",24,in);
}

void fit_set_fish_voltage(double in)
{
edit_file("fish.inp","",14,in);
}

void fit_set_externalv_ce(double in)
{
printf("setting %le\n",in);
edit_file("ce.inp","",6,in);
}

void fit_set_dt(double in)
{
edit_file("tpc.inp","",6,in);
}

void fit_set_rep_rate(int in)
{
edit_file_int("tpc.inp","",20,in);
}

void fit_set_pulse_len(int in)
{
edit_file_int("tpc.inp","",18,in);
}

void fit_set_tpc_stop(double in)
{
edit_file("tpc.inp","",2,in);
}

void fit_set_tof_stop(double in)
{
edit_file("tof.inp","",2,in);
}

void fit_set_ce_stop(double in)
{
edit_file("ce.inp","",4,in);
}

double fit_run_sims(struct fitvars *fitconfig,struct server *myserver)
{

double error=0.0;
double Vstart=0.0;
double Vstop=0.0;
double Vstep=0.0;
char tempsim[1000];
char tempexp[1000];
char command[1000];
char tempstring[1000];

char config_files[3000];

  DIR           *d;
  struct dirent *dir;
  d = opendir(".");
strcpy(config_files,"");
  if (d)
  {
    while ((dir = readdir(d)) != NULL)
    {
     if (strcmp_end(dir->d_name,".inp")==0)
	{
		strcat(config_files,dir->d_name);
		strcat(config_files," ");
	}
    }

    closedir(d);
  }
printf("copy files= %s\n",config_files);

strcpy(config_files,config_files);
double voltage;
int i;

gen_dos_fd_gaus_fd();

//getchar();
//printf("about to run for dark\n");
mkdir("./sim", S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);

if (fitconfig->fit_dark_iv==TRUE)
{
	printf("Running dark jv\n");
	sprintf(tempsim,"./sim/iv");
	sprintf(tempexp,"./exp/iv");

	mkdir(tempsim, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
	strcat(tempexp,"/");
	strcat(tempsim,"/");

	FILE *info=fopena(tempexp,"info.inp","r");
	if (info == NULL)
	{
		printf("Error with %sinfo.inp\n",tempexp);
		exit(0);
	}

	fscanf(info,"%s",tempstring);
	fscanf(info,"%le",&Vstart);
	fscanf(info,"%s",tempstring);
	fscanf(info,"%le",&Vstop);
	fscanf(info,"%s",tempstring);
	fscanf(info,"%le",&Vstep);
	fscanf(info,"%s",tempstring);

	if (strcmp("#end",tempstring)!=0)
	{
		printf("Problem with %s/info.inp\n",tempexp);
		exit(0);
	}
	fclose(info);

	fit_set_light(0.0);
	fit_set_volt_step(Vstep);				//change step
	fit_set_volt_start(Vstart);				//change vstart
	fit_set_volt_stop(Vstop);				//change vstop
	fit_set_sim_mode(0);					//set simulation mode
	sprintf(command,"cp %s %s",config_files,tempsim);
	system(command);
	server_add_job(myserver,tempsim);
}



if (fitconfig->fit_light_iv==TRUE)
{
	sprintf(tempsim,"./sim/ivlight");
	sprintf(tempexp,"./exp/ivlight");
	mkdir(tempsim, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
	strcat(tempexp,"/");
	strcat(tempsim,"/");

	FILE *info=fopena(tempexp,"info.inp","r");

	fscanf(info,"%s",tempstring);
	fscanf(info,"%le",&Vstart);
	fscanf(info,"%s",tempstring);
	fscanf(info,"%le",&Vstop);
	fscanf(info,"%s",tempstring);
	fscanf(info,"%le",&Vstep);
	fscanf(info,"%s",tempstring);

	if (strcmp("#end",tempstring)!=0)
	{
		printf("Problem with info.inp in %s!\n",tempexp);
		exit(0);
	}
	fclose(info);

	fit_set_light(1000.0);					//light on
	fit_set_volt_step(Vstep);				//change step
	fit_set_volt_start(Vstart);				//change vstart
	fit_set_volt_stop(Vstop);				//change vstop
	fit_set_sim_mode(0);					//set simulation mode

	sprintf(command,"cp %s %s",config_files,tempsim);
	system(command);
	server_add_job(myserver,tempsim);
	fit_set_light(0.0);					//light off
}

if (fitconfig->fit_light2_iv==TRUE)
{
	sprintf(tempsim,"./sim/ivlight2");
	mkdir(tempsim, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
	strcat(tempsim,"/");
	fit_set_light(fitconfig->light2_power);					//light on
	fit_set_volt_step(0.01);				//change step
	fit_set_volt_start(0.3);				//change vstart
	fit_set_sim_mode(0);					//set simulation mode

	sprintf(command,"cp %s %s",config_files,tempsim);
	system(command);
	server_add_job(myserver,tempsim);
	fit_set_light(0.0);					//light off
}

for (i=0;i<fitconfig->fish_max;i++)
{

	printf("Running time fish %d/%d\n",i,fitconfig->fish_max);
	sprintf(tempsim,"./sim/fish%d",i);
	sprintf(tempexp,"./exp/fish%d",i);
	mkdir(tempsim, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);

	strcat(tempexp,"/");
	strcat(tempsim,"/");

	FILE *info=fopena(tempexp,"info.inp","r");
	fscanf(info,"%s",tempstring);
	fscanf(info,"%le",&voltage);
	fscanf(info,"%s",tempstring);

	if (strcmp("#end",tempstring)!=0)
	{
		printf("Problem with %sinfo.inp\n",tempexp);
		exit(0);
	}
	fclose(info);

	printf("Doing fish %d %lf\n",i,voltage);

	fit_set_light(0.0);					//light bias
	fit_set_sim_mode(17);					//set simulation mode
	fit_set_fish_voltage(voltage);				//set externalv

	sprintf(command,"cp %s %s",config_files,tempsim);
	system(command);
	server_add_job(myserver,tempsim);
}

for (i=0;i<fitconfig->tpcmax;i++)
{

	double suns=0.0;
	double tpc_stop=0.0;
	double dt=0.0;
	printf("Running time tpc %d/%d\n",i,fitconfig->tpcmax);
	sprintf(tempsim,"./sim/tpc%d",i);
	sprintf(tempexp,"./exp/tpc%d",i);
	mkdir(tempsim, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);

	strcat(tempexp,"/");
	strcat(tempsim,"/");

	FILE *info=fopena(tempexp,"info.inp","r");
	fscanf(info,"%s",tempstring);
	fscanf(info,"%le",&suns);
	fscanf(info,"%s",tempstring);
	fscanf(info,"%le",&voltage);
	fscanf(info,"%s",tempstring);
	fscanf(info,"%le",&tpc_stop);
	fscanf(info,"%s",tempstring);
	fscanf(info,"%le",&dt);
	fscanf(info,"%s",tempstring);

	if (strcmp("#end",tempstring)!=0)
	{
		printf("Problem with %sinfo.inp\n",tempexp);
		exit(0);
	}
	fclose(info);

	printf("Doing tpc %d %lf\n",i,voltage);

	fit_set_light(suns);					//light bias
	fit_set_sim_mode(4);					//set simulation mode
	fit_set_tpc_externalv(voltage);				//set externalv

	fit_set_rep_rate(150);					//set rep rate
	fit_set_dt(dt);				//set dt
	fit_set_tpc_stop(tpc_stop);				//set tpc timemax
	fit_set_pulse_len(2);					//pulse length
	sprintf(command,"cp %s %s",config_files,tempsim);
	system(command);
	server_add_job(myserver,tempsim);
}


for (i=0;i<fitconfig->tofmax;i++)
{

	double suns=0.0;
	double tof_stop=0.0;
	printf("Running time tof %d/%d\n",i,fitconfig->tofmax);
	sprintf(tempsim,"./sim/tof%d",i);
	sprintf(tempexp,"./exp/tof%d",i);
	mkdir(tempsim, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);

	strcat(tempexp,"/");
	strcat(tempsim,"/");

	FILE *info=fopena(tempexp,"info.inp","r");
	fscanf(info,"%s",tempstring);
	fscanf(info,"%le",&suns);
	fscanf(info,"%s",tempstring);
	fscanf(info,"%le",&voltage);
	fscanf(info,"%s",tempstring);
	fscanf(info,"%le",&tof_stop);
	fscanf(info,"%s",tempstring);

	if (strcmp("#end",tempstring)!=0)
	{
		printf("Problem with info.inp in %s!\n",tempexp);
		exit(0);
	}
	fclose(info);

	printf("Doing tof %d %lf\n",i,voltage);

	fit_set_light(suns);					//light bias
	fit_set_sim_mode(11);					//set simulation mode
	fit_set_external_volt(voltage);				//set externalv

	fit_set_rep_rate(150);					//set rep rate
	fit_set_dt(2.000000e-7);				//set dt
	fit_set_tof_stop(tof_stop);				//set tof timemax
	fit_set_pulse_len(2);					//pulse length
	sprintf(command,"cp %s %s",config_files,tempsim);
	system(command);
	server_add_job(myserver,tempsim);
}

for (i=0;i<fitconfig->cemax;i++)
{

	double suns=0.0;
	double tof_stop=0.0;
	printf("Running ce%d/%d\n",i,fitconfig->tofmax);
	sprintf(tempsim,"./sim/ce%d",i);
	sprintf(tempexp,"./exp/ce%d",i);
	mkdir(tempsim, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);

	strcat(tempexp,"/");
	strcat(tempsim,"/");

	FILE *info=fopena(tempexp,"info.inp","r");
	fscanf(info,"%s",tempstring);
	fscanf(info,"%le",&suns);
	fscanf(info,"%s",tempstring);
	fscanf(info,"%le",&voltage);
	fscanf(info,"%s",tempstring);
	fscanf(info,"%le",&tof_stop);
	fscanf(info,"%s",tempstring);

	if (strcmp("#end",tempstring)!=0)
	{
		printf("Problem with info.inp in %s!\n",tempexp);
		exit(0);
	}
	fclose(info);

	printf("Doing ce %d %lf\n",i,voltage);

	fit_set_light(suns);					//light bias
	fit_set_sim_mode(2);					//set simulation mode
	fit_set_externalv_ce(voltage);				//set externalv
	fit_set_ce_stop(tof_stop);				//set tof timemax

	sprintf(command,"cp %s %s",config_files,tempsim);
	system(command);
	server_add_job(myserver,tempsim);
}

int ce_dark_max=0;
if (fitconfig->dotimece==TRUE)
{
	struct istruct exp;
	inter_load_a(&exp,"./exp/cedata.dat",0,1);
	double start=exp.x[0];
	double stop=exp.x[exp.len-1];
	double steps=3.0;
	double step=(stop-start)/steps;
	//printf("%e %e\n",start,stop);
	//getchar();
	inter_free(&exp);
	double V=start;

	do
	{
	printf("Running time ce %d\n",ce_dark_max);
	sprintf(tempsim,"./sim/ce%d",ce_dark_max);
	sprintf(tempexp,"./exp/ce%d",ce_dark_max);
	mkdir(tempsim, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);

	strcat(tempexp,"/");
	strcat(tempsim,"/");


	fit_set_light(0.0);					//light off
	fit_set_volt_start(0.05);				//change vstart
	fit_set_sim_mode(2);					//set simulation mode
	fit_set_dt(2.4000e-06);					//set dt
	fit_set_rep_rate(25);					//set rep rate
	fit_set_pulse_len(4);					//pulse length
	fit_set_external_volt(V);				//set external voltage
	sprintf(command,"cp %s %s",config_files,tempsim);
	system(command);

	server_add_job(myserver,tempsim);

	V+=step;
	ce_dark_max++;
	}while(ce_dark_max<=steps);


	//run_simulation(tempsim);
}


int ce_light_max=0;
if (fitconfig->dotimecelight==TRUE)
{
	struct istruct exp;
	inter_load_a(&exp,"./exp/celight.dat",0,1);
	double start=exp.x[0];
	double stop=exp.x[exp.len-1];
	double steps=3.0;
	double step=(stop-start)/steps;
	//printf("%e %e\n",start,stop);
	//getchar();
	inter_free(&exp);
	double V=start;

	do
	{
	printf("Running time ce light %d\n",ce_light_max);
	sprintf(tempsim,"./sim/celight%d",ce_light_max);
	sprintf(tempexp,"./exp/celight%d",ce_light_max);
	mkdir(tempsim, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);

	strcat(tempexp,"/");
	strcat(tempsim,"/");


	fit_set_light(1000.0);					//One sun
	fit_set_volt_start(0.05);		//change vstart
	fit_set_sim_mode(2);					//set simulation mode
	fit_set_dt(2.4000e-06);					//set dt
	fit_set_rep_rate(25);					//set rep rate
	fit_set_pulse_len(4);					//pulse length
	fit_set_external_volt(V);				//set external voltage
	sprintf(command,"cp %s %s",config_files,tempsim);
	system(command);

	server_add_job(myserver,tempsim);

	V+=step;
	ce_light_max++;
	}while(ce_light_max<=steps);


	//run_simulation(tempsim);
}
//fitconfig->tpclightmax

print_jobs(myserver);
server_run_jobs(myserver);

//printf("here\n");
//getchar();
if (fitconfig->dotimece==TRUE)
{
	mkdir("./sim/ce", S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
	FILE *r;
	r=fopen("./sim/ce/realce.dat","w");
	fclose(r);
	double nread=0.0;
	double V=0.0;
	for (i=0;i<ce_dark_max;i++)
	{
		sprintf(tempsim,"./sim/ce%d/cej.dat",i);
		r=fopen(tempsim,"r");
		fscanf(r,"%le %le",&V,&nread);
		fclose(r);
		r=fopen("./sim/ce/realce.dat","a");
		fprintf(r,"%le %le\n",V,nread);
		fclose(r);
	}
}

if (fitconfig->dotimecelight==TRUE)
{
	mkdir("./sim/celight", S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
	FILE *r;
	r=fopen("./sim/celight/realce.dat","w");
	fclose(r);
	double nread=0.0;
	double V=0.0;
	for (i=0;i<ce_light_max;i++)
	{
		sprintf(tempsim,"./sim/celight%d/cej.dat",i);
		r=fopen(tempsim,"r");
		fscanf(r,"%le %le",&V,&nread);
		fclose(r);
		r=fopen("./sim/celight/realce.dat","a");
		fprintf(r,"%le %le\n",V,nread);
		fclose(r);
	}
}
error=get_all_error(fitconfig);
printf("Jobs/s=%le ODEs/s=%le\n",server_get_jobs_per_s(),server_get_odes_per_s());
return error;
}

